/* #define DEBUG */
/* 
 * tkForeignWin.c --
 *
 *	This module implements "foreignwin" widgets.  A "foreignwin" is
 *	a widget that contains a window from another application.
 */

#include "tk.h"
#include "foreignwin.h"
/*
 * A data structure of the following type is kept for each foreignwin
 * widget managed by this file:
 */

typedef struct {
    Tk_Window tkwin;		/* Window that embodies the foreignwin.  NULL
				 * means window has been deleted but
				 * widget record hasn't been cleaned up yet. */
    Display *display;		/* X's token for the window's display. */
    Tcl_Interp *interp;		/* Interpreter associated with widget. */
    Tcl_Command widgetCmd;	/* Token for foreignwin's widget command. */
    Window foreignWinId;	/* X window Id of foreign window. */
    int width, height;		/* Desired width and height for the window;
				 * <= 0 means use the foreign window's
				 * width and height. */
    int followSize;		/* One means that the Tk window should
				 * follow the size of the foreign window,
				 * i.e. handle ConfigureRequest events by
				 * automatically granting them. Zero
				 * means don't do it; let Tk bindings
				 * handle the resizing. */
    int reqWidth, reqHeight;	/* Width and height, as requested either
				 * from the foreign window or from a
				 * "configure" widget command. */
    int updatePending;		/* Non-zero means a call to ForeignWinDisplay
				 * has already been scheduled. */
    Tk_OptionTable optionTable;
} ForeignWin;

/*
 * Information used for argv parsing.
 */

static Tk_OptionSpec configSpecs[] = {
    {TK_OPTION_INT, "-followsize", "followsize", "FollowSize",
	"1", -1, Tk_Offset(ForeignWin, followSize), 0},
    {TK_OPTION_PIXELS, "-height", "height", "Height",
	"0", -1, Tk_Offset(ForeignWin, height), 0},
    {TK_OPTION_PIXELS, "-width", "width", "Width",
	"0", -1, Tk_Offset(ForeignWin, width), 0},
    {TK_OPTION_END}
};

/*
 * Forward declarations for procedures defined later in this file:
 */

int			ForeignWinCmd _ANSI_ARGS_((ClientData clientData,
			    Tcl_Interp *interp, int objc, struct Tcl_Obj *CONST objv[]));
static void		ForeignWinCmdDeletedProc _ANSI_ARGS_((
			    ClientData clientData));
static int		ForeignWinConfigure _ANSI_ARGS_((Tcl_Interp *interp,
			    ForeignWin *foreignwinPtr, int objc, struct Tcl_Obj *CONST objv[],
			    int flags));
static void		ForeignWinDestroy _ANSI_ARGS_((char *memPtr));
static void		ForeignWinDisplay _ANSI_ARGS_((ClientData clientData));
static void		ForeignWinEventProc _ANSI_ARGS_((ClientData clientData,
			    XEvent *eventPtr));
static int		ForeignWinWidgetCmd _ANSI_ARGS_((ClientData clientData,
			    Tcl_Interp *, int objc, struct Tcl_Obj *CONST objv[]));

/* Error handler, used below. */

#define X_ERROR_LENGTH  300

static int
eHand(clientData, errEventPtr)
    ClientData clientData;
    XErrorEvent *errEventPtr;
{
    char *xErrorMsg;

    xErrorMsg = (char *) clientData;
    XGetErrorText(errEventPtr->display, errEventPtr->error_code,
		xErrorMsg, X_ERROR_LENGTH);
    return 0;
}

/* Table of foreign X window IDs. */

static Tcl_HashTable foreignWinTable;
static int tableInitialized = 0;

/* Get control of Events for foreign (non-Tk) windows. */
static int TkEventHandler( ClientData clientData, XEvent *eventPtr) {
    Tcl_HashEntry *hashEntryPtr;
    
    ForeignWin *foreignwinPtr=(void*)clientData;
#ifdef DEBUG
    printf("Got external event %d for %x\n", eventPtr->type, foreignwinPtr->foreignWinId);
#endif
    if (eventPtr->xany.window != foreignwinPtr->foreignWinId) {
      return 0;
    }
    if (eventPtr->type == DestroyNotify) {
      ForeignWinEventProc(clientData, eventPtr);
    }
    return 1;
}

/* This doesn't really belong here! */
static int
Tkwm_XErrorHandler (clientData, errEventPtr)
    ClientData clientData;    /* Pointer to flag we set       */
    XErrorEvent *errEventPtr; /* X error info                 */
{
    int *error;

    error = (int *) clientData;
    *error = 1;
    return 0;
}

static int
Tkwm_GetWindowAttributes(display,win,attrPtr)
    Display *display;
    Window win;
    XWindowAttributes *attrPtr;
{
    Tk_ErrorHandler handler;
    Status status;
    int gotXError;

    gotXError = 0;
    handler = Tk_CreateErrorHandler(display,
              -1, -1, -1, Tkwm_XErrorHandler,
              (ClientData) (&gotXError));
    status = XGetWindowAttributes(display,win,attrPtr);
    Tk_DeleteErrorHandler(handler);
    XSync(display, False);
    return status;
}


static int
Tkwm_GetDrawableId(interp, str, drawablePtr)
    Tcl_Interp *interp; /* Error messages go here */
    char *str;          /* String to convert to drawable ID */
    Drawable *drawablePtr;  /* Drawable ID goes in here */
{
    int i, result;

    i = (int) None;
    result = Tcl_GetInt(interp,str,&i);
    *drawablePtr = (Drawable) i;
    return result;
}

static int
Tkwm_IsValidWindow(display, window)
    Display *display;
    Window window;
{
    XWindowAttributes attr;

    return(Tkwm_GetWindowAttributes(display, window, &attr));
}

static  int
Tkwm_GetAndCheckWindowId(interp, display, str, windowPtr)
    Tcl_Interp *interp;
    Display *display;
    char *str;
    Window *windowPtr;
{
    int result;

    /* Get the window ID, and make sure that it exists. */

    result = Tkwm_GetDrawableId(interp,str,(Drawable *) windowPtr);
    if (result != TCL_OK) {
      return TCL_ERROR;
    }

    if (Tkwm_IsValidWindow(display, *windowPtr)) {
      return TCL_OK;
    }

    Tcl_AppendResult(interp, "No window ", str, " on display", (char *) 0);
    return TCL_ERROR;
}

/*
 *@@@ NOTE: the table really should be indexed by
 *@@@ BOTH the window ID AND the display.
 */

/*
 *--------------------------------------------------------------
 *
 * Tk_XIdIsForeignWin --
 *
 *	See if a given X window is the slave window of a
 *	foreignwin widget.
 *	
 *
 * Results:
 *	Returns 1 if it is, 0 if it isn't.
 *
 * Side effects:
 *	None
 *
 *--------------------------------------------------------------
 */

int
Tk_XIdIsForeignWin(XId)
    Window XId;
{
    Tcl_HashEntry *hashEntryPtr;

    /*
     * If the table's not initialized, then the window can't
     * be the ID of a foreignwin slave, because there aren't any.
     */

    if (!tableInitialized) {
	return 0;
    }

    /* See if there's an entry in the hash table. */

    hashEntryPtr = Tcl_FindHashEntry(&foreignWinTable, (char *) XId);
    return (hashEntryPtr != NULL);
}

/*
 *--------------------------------------------------------------
 *
 * ForeignWinCmd --
 *
 *	This procedure is invoked to process the "foreignwin" Tcl
 *	command.  It creates a new "foreignwin" widget.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	A new widget is created and configured.
 *
 *--------------------------------------------------------------
 */

int
Tk_ForeignWinCmd(clientData, interp, objc, objv)
    ClientData clientData;	/* Main window associated with
				 * interpreter. */
    Tcl_Interp *interp;		/* Current interpreter. */
    int objc;			/* Number of arguments. */
    struct Tcl_Obj *CONST objv[];	/* Argument strings. */
{
    Tk_Window main = (Tk_Window) clientData;
    ForeignWin *foreignwinPtr;
    Tk_Window tkwin;
 /*   TkWindow *winPtr; */
    Window foreignXId;
    char junkStr[32];
    Tk_ErrorHandler handler;
    int xErrorMsg[X_ERROR_LENGTH + 1];
    Tcl_HashEntry *hashEntryPtr;
    int unused;


    if (objc < 3) {
	Tcl_WrongNumArgs(interp, 1, objv, " pathName XWinId ?options?");
	return TCL_ERROR;
    }

    /* Make sure things are initialized. */

    if (!tableInitialized) {
	Tcl_InitHashTable(&foreignWinTable, TCL_ONE_WORD_KEYS);
	tableInitialized = 1;
    }

    /*
     * Get the X window ID of the foreign window, make sure that it
     * exists, and that it's not from this application.
     */

    if (Tkwm_GetAndCheckWindowId(interp,
		Tk_Display(main), Tcl_GetString(objv[2]), &foreignXId) != TCL_OK) {
        return(TCL_ERROR);
    }

#ifdef DEBUG
    printf("New Foreignwin: %x\n", foreignXId);
#endif
    /* Make sure that foreignXId isn't already a Tk window. */
    /*@@@ Do we want to do this, or just let the window be reparented? @@@*/

    tkwin = Tk_IdToWindow(Tk_Display(main), foreignXId);
    if (tkwin != NULL) {
        Tcl_ResetResult(interp);
        sprintf(junkStr,"0x%x", (unsigned int) foreignXId);
        Tcl_AppendResult(interp, "X Window ", junkStr,
                " is already a Tk widget (",
                Tk_PathName(tkwin), ")", (char *) 0);
        return(TCL_ERROR);
    }

    /* See if foreignXId has already been used. */

    hashEntryPtr = Tcl_CreateHashEntry(&foreignWinTable,
					(char *) foreignXId, &unused);
    if (!unused) {
	Tcl_ResetResult(interp);
	sprintf(junkStr,"0x%x", (unsigned int) foreignXId);
        Tcl_AppendResult(interp, "X Window ", junkStr,
                " is already used in a foreignwin widget", (char *) 0);
        return(TCL_ERROR);
    }

    /*
     * Create the window we're going to reparent into.
     */

    tkwin = Tk_CreateWindowFromPath(interp, main, Tcl_GetString(objv[1]), (char *) NULL);
    if (tkwin == NULL) {
	return TCL_ERROR;
    }
    Tk_SetClass(tkwin, "ForeignWin");

    /*
     * Reparent the foreign window into the Tk window.
     */

    Tk_MakeWindowExist(tkwin);
    xErrorMsg[0] = '\0';
    handler = Tk_CreateErrorHandler(Tk_Display(tkwin), -1, -1, -1,
		eHand, (ClientData) xErrorMsg);
    XReparentWindow(Tk_Display(tkwin),
                foreignXId, Tk_WindowId(tkwin), 0, 0);
    Tk_DeleteErrorHandler(handler);
    XSync(Tk_Display(tkwin), False);
    if(xErrorMsg[0] != '\0') {
	Tk_DestroyWindow(tkwin);
        Tcl_AppendResult(interp, xErrorMsg, (char *) 0);
        return(TCL_ERROR);
    }

    /*
     * Make sure the foreign window is mapped.
     */

    XMapWindow(Tk_Display(tkwin), foreignXId);

    /*
     * Allocate and initialize the widget record.
     */

    foreignwinPtr = (ForeignWin *) ckalloc(sizeof(ForeignWin));
    foreignwinPtr->tkwin = tkwin;
    foreignwinPtr->display = Tk_Display(tkwin);
    foreignwinPtr->interp = interp;
    foreignwinPtr->widgetCmd = Tcl_CreateObjCommand(interp,
	    Tk_PathName(foreignwinPtr->tkwin), ForeignWinWidgetCmd,
	    (ClientData) foreignwinPtr, ForeignWinCmdDeletedProc);
    foreignwinPtr->width = 0;
    foreignwinPtr->height = 0;
    foreignwinPtr->reqWidth = -1;
    foreignwinPtr->reqHeight = -1;
    foreignwinPtr->updatePending = 0;
    foreignwinPtr->followSize = 1;
    foreignwinPtr->foreignWinId = foreignXId;
    foreignwinPtr->optionTable=Tk_CreateOptionTable(interp, configSpecs);

    /* Add foreign window to table. */

    Tcl_SetHashValue(hashEntryPtr, (ClientData) foreignwinPtr);

    /* Make sure we get PropertyNotify events for the foreign window. */

    XSelectInput(foreignwinPtr->display, foreignXId, PropertyChangeMask|StructureNotifyMask);

    Tk_CreateGenericHandler(TkEventHandler, foreignwinPtr);
    /* Make sure we get SubstructureNotify  (etc.) events. */

/*    winPtr = (TkWindow *) tkwin;
    winPtr->atts.event_mask |= ExposureMask
			|SubstructureRedirectMask
			|StructureNotifyMask|SubstructureNotifyMask
			|PropertyChangeMask|ResizeRedirectMask;
    XSelectInput(winPtr->display, winPtr->window, winPtr->atts.event_mask); */

    if (Tk_InitOptions(interp, (char *) foreignwinPtr, foreignwinPtr->optionTable, tkwin)
		                != TCL_OK) {
	Tk_DestroyWindow(foreignwinPtr->tkwin);
	return TCL_ERROR;

    }
    Tk_CreateEventHandler(foreignwinPtr->tkwin,
			ExposureMask|StructureNotifyMask|SubstructureNotifyMask|SubstructureRedirectMask|PropertyChangeMask|ResizeRedirectMask,
			ForeignWinEventProc, (ClientData) foreignwinPtr);
    if (ForeignWinConfigure(interp, foreignwinPtr, objc-3, objv+3, 0) != TCL_OK) {
	Tk_DestroyWindow(foreignwinPtr->tkwin);
	return TCL_ERROR;
    }

    interp->result = Tk_PathName(foreignwinPtr->tkwin);
    return TCL_OK;
}

/*
 *--------------------------------------------------------------
 *
 * ForeignWinWidgetCmd --
 *
 *	This procedure is invoked to process the Tcl command
 *	that corresponds to a widget managed by this module.
 *	See the user documentation for details on what it does.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	See the user documentation.
 *
 *--------------------------------------------------------------
 */

static int
ForeignWinWidgetCmd(clientData, interp, objc, objv)
    ClientData clientData;		/* Information about foreignwin widget. */
    Tcl_Interp *interp;			/* Current interpreter. */
    int objc;				/* Number of arguments. */
    struct Tcl_Obj *CONST objv[];	/* Argument strings. */
{
    ForeignWin *foreignwinPtr = (ForeignWin *) clientData;
    int result = TCL_OK;
    size_t length;
    char c;
    char junkStr[32];
    char *arg;
    Tcl_Obj *rObj;

    if (objc < 2) {
	Tcl_WrongNumArgs(interp, 1, objv, "option ?arg arg ...?");
	return TCL_ERROR;
    }
    arg = Tcl_GetStringFromObj(objv[1], &length);
    Tcl_Preserve((ClientData) foreignwinPtr);
    c = arg[0];
    if ((c == 'c') && (strncmp(arg, "cget", length) == 0)
	    && (length >= 2)) {
	if (objc != 3) {
	    Tcl_WrongNumArgs(interp, 2, objv," cget option");
	    goto error;
	}
	rObj = Tk_GetOptionValue(interp, (char *) foreignwinPtr,
		foreignwinPtr->optionTable, objv[2], foreignwinPtr->tkwin);
	goto objret;
    } else if ((c == 'c') && (strncmp(arg, "configure", length) == 0)
	    && (length >= 2)) {
	if (objc == 2) {
	    rObj = Tk_GetOptionInfo(interp, (char *) foreignwinPtr,
		foreignwinPtr->optionTable, NULL, foreignwinPtr->tkwin);
	    goto objret;
	} else if (objc == 3) {
	    rObj = Tk_GetOptionInfo(interp, (char *) foreignwinPtr,
		foreignwinPtr->optionTable, objv[2], foreignwinPtr->tkwin);
	    goto objret;
	} else {
	    result = ForeignWinConfigure(interp, foreignwinPtr, objc-2, objv+2,
		    TK_CONFIG_ARGV_ONLY);
	}
    } else if ((c == 'x') && (strcmp(arg, "xid") == 0)) {
	sprintf(junkStr, "0x%x", (unsigned int) (foreignwinPtr->foreignWinId));
	Tcl_SetResult(interp, junkStr, TCL_VOLATILE);
	result = TCL_OK;
    } else {
	Tcl_AppendResult(interp, "bad option \"", arg,
		"\": must be cget, configure, or xid",
		(char *) NULL);
	goto error;
    }
    done:
    if (!foreignwinPtr->updatePending) {
	Tcl_DoWhenIdle(ForeignWinDisplay, (ClientData) foreignwinPtr);
	foreignwinPtr->updatePending = 1;
    }
    Tcl_Release((ClientData) foreignwinPtr);
    return result;

    error:
    Tcl_Release((ClientData) foreignwinPtr);
    return TCL_ERROR;
    objret:
    if (!rObj) {
        result = TCL_ERROR;
    } else {
        result = TCL_OK;
        Tcl_SetObjResult(interp, rObj);
      }
    goto done;
}

/*
 *----------------------------------------------------------------------
 *
 * ForeignWinConfigure --
 *
 *	This procedure is called to process an argv/argc list in
 *	conjunction with the Tk option database to configure (or
 *	reconfigure) a foreignwin widget.
 *
 * Results:
 *	The return value is a standard Tcl result.  If TCL_ERROR is
 *	returned, then interp->result contains an error message.
 *
 * Side effects:
 *	Configuration information, such as colors, border width,
 *	etc. get set for foreignwinPtr;  old resources get freed,
 *	if there were any.
 *
 *----------------------------------------------------------------------
 */

static int
ForeignWinConfigure(interp, foreignwinPtr, objc, objv, flags)
    Tcl_Interp *interp;			/* Used for error reporting. */
    ForeignWin *foreignwinPtr;		/* Information about widget. */
    int objc;				/* Number of valid entries in objv. */
    struct Tcl_Obj *CONST objv[];	/* Arguments. */
    int flags;				/* Flags to pass to
					 * Tk_ConfigureWidget. */
{
    int w,h;
    XWindowAttributes winAtts;
    Status status;

    if (Tk_SetOptions(interp, (char *)foreignwinPtr, foreignwinPtr->optionTable,
		    objc, objv, foreignwinPtr->tkwin, 0, 0) != TCL_OK) {
	return TCL_ERROR;
    }

    /*
     * Register the desired geometry for the window.  Then arrange for
     * the window to be redisplayed.
     */

    w = foreignwinPtr->width;
    h = foreignwinPtr->height;
    if (w <= 0 || h <= 0) {
	status = XGetWindowAttributes(Tk_Display(foreignwinPtr->tkwin),
				foreignwinPtr->foreignWinId, &winAtts);
	if (status == 0) {
	   return TCL_ERROR;
	}
	if (w <= 0) {
	    w = winAtts.width;
	}
	if (h <= 0) {
	    h = winAtts.height;
	}
    }

#ifdef DEBUG
    Tk_GeometryRequest(foreignwinPtr->tkwin, w, h);
    XResizeWindow(Tk_Display(foreignwinPtr->tkwin),
		 foreignwinPtr->foreignWinId, w, h);
#else
    foreignwinPtr->reqWidth = w;
    foreignwinPtr->reqHeight = h;
#endif


    if (!foreignwinPtr->updatePending) {
	Tcl_DoWhenIdle(ForeignWinDisplay, (ClientData) foreignwinPtr);
	foreignwinPtr->updatePending = 1;
    }
    return TCL_OK;
}

/*
 *--------------------------------------------------------------
 *
 * ForeignWinEventProc --
 *
 *	This procedure is invoked by the Tk dispatcher for various
 *	events on foreignwins.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	When the window gets deleted, internal structures get
 *	cleaned up.  When it gets exposed, it is redisplayed.
 *
 *--------------------------------------------------------------
 */

int foo() {
  return 1;
}
static void
ForeignWinEventProc(clientData, eventPtr)
    ClientData clientData;	/* Information about window. */
    XEvent *eventPtr;		/* Information about event. */
{
    Window foreignXId;
    XWindowChanges changes;
    Tcl_HashEntry *hashEntryPtr;
    Tk_ErrorHandler handler;
    int gotXError;

    ForeignWin *foreignwinPtr = (ForeignWin *) clientData;
    foreignXId = foreignwinPtr->foreignWinId;

#ifdef DEBUG
    fprintf(stderr, "Handling event; type = %1d, window = 0x%08x\n",
					eventPtr->type, eventPtr->xany.window);
#endif
    if (eventPtr->type == Expose) {
	/*
	 * If the widget window was exposed, make sure
	 * that the foreign window gets an expose event.
	 */

	/*@@@ Doesn't look like we need to do this. @@@*/
    } else if (eventPtr->type == ConfigureNotify) {
#ifdef DEBUG
	fprintf(stderr, "Handling ConfigureNotify event (win = 0x%08x)\n",
			eventPtr->xconfigure.window);
#endif

	if (eventPtr->xconfigure.window == Tk_WindowId(foreignwinPtr->tkwin)) {
	    /*
	     * Make sure the foreign window stays the same size
	     * as the Tk window.
	     */

#ifdef DEBUG
	    fprintf(stderr,
		"Setting req. width and height; updatePending = %1d\n",
						foreignwinPtr->updatePending);
	    fprintf(stderr, "	rw = %1d, rh = %1d\n",
			eventPtr->xconfigure.width,
			eventPtr->xconfigure.height);
	    fprintf(stderr, "	tkrw = %1d, tkrw = %1d\n",
			Tk_ReqWidth(foreignwinPtr->tkwin),
			Tk_ReqHeight(foreignwinPtr->tkwin));
#endif
	    foreignwinPtr->reqWidth = eventPtr->xconfigure.width;
	    foreignwinPtr->reqHeight = eventPtr->xconfigure.height;
	    if (!foreignwinPtr->updatePending) {
		foreignwinPtr->updatePending = 1;
#ifdef DEBUG
		fprintf(stderr, "	Calling Tcl_DoWhenIdle\n");
#endif
		Tcl_DoWhenIdle(ForeignWinDisplay, (ClientData) foreignwinPtr);
#ifdef DEBUG
		fprintf(stderr, "	Returned from Tcl_DoWhenIdle\n");
#endif
	    }
	}
#if 1
	else if (eventPtr->xconfigure.window == foreignwinPtr->foreignWinId
		&& foreignwinPtr->followSize) {
	    foreignwinPtr->reqWidth = eventPtr->xconfigure.width;
	    foreignwinPtr->reqHeight = eventPtr->xconfigure.height;
	    if (!foreignwinPtr->updatePending) {
		foreignwinPtr->updatePending = 1;
		Tcl_DoWhenIdle(ForeignWinDisplay, (ClientData) foreignwinPtr);
	    }
	}
#endif

	/*@@@ This is probably not needed. @@@*/
	if (!foreignwinPtr->updatePending) {
	    foreignwinPtr->updatePending = 1;
#ifdef DEBUG
	    fprintf(stderr, "	Calling Tcl_DoWhenIdle (2)\n");
#endif
	    Tcl_DoWhenIdle(ForeignWinDisplay, (ClientData) foreignwinPtr);
#ifdef DEBUG
	    fprintf(stderr, "	Returning from Tcl_DoWhenIdle (2)\n");
#endif
	}
    } else if (eventPtr->type == ConfigureRequest) {
	/*
	 * If we're supposed to follow the size of the foreign
	 * window, resize the Tk window accordingly.
	 */

#ifdef DEBUG
	fprintf(stderr, "Handling ConfigureRequest event\n");
#endif
	if (eventPtr->xconfigurerequest.window == foreignwinPtr->foreignWinId
		&& eventPtr->xconfigurerequest.parent
				== Tk_WindowId(foreignwinPtr->tkwin)
		&& foreignwinPtr->followSize) {
	    foreignwinPtr->reqWidth = eventPtr->xconfigurerequest.width;
	    foreignwinPtr->reqHeight = eventPtr->xconfigurerequest.height;
	    if (!foreignwinPtr->updatePending) {
		foreignwinPtr->updatePending = 1;
		Tcl_DoWhenIdle(ForeignWinDisplay, (ClientData) foreignwinPtr);
	    }
	}
    } else if (eventPtr->type == DestroyNotify) {
	/* Gotta handle destruction of widget AND foreign window. */
	    foo();

	if(eventPtr->xdestroywindow.window == foreignwinPtr->foreignWinId) {
	    /* Foreign window destroyed; destroy the Tk window. */

	    hashEntryPtr = Tcl_FindHashEntry(&foreignWinTable,
				(char *) (foreignwinPtr->foreignWinId));
	    if (hashEntryPtr != NULL) {
		Tcl_DeleteHashEntry(hashEntryPtr);
	    }
	    foreignwinPtr->foreignWinId = None;
            Tcl_Preserve((ClientData) foreignwinPtr);
	    Tk_DestroyWindow(foreignwinPtr->tkwin);
	    foreignwinPtr->tkwin = NULL;
            Tcl_Release((ClientData) foreignwinPtr);
	} else {
	    if (foreignwinPtr->tkwin != NULL) {
		if (foreignwinPtr->foreignWinId != None) {
		    gotXError = 0;
		    handler = Tk_CreateErrorHandler(Tk_Display(foreignwinPtr->tkwin),
				 -1, -1, -1,
				Tkwm_XErrorHandler, (ClientData) (&gotXError));
		    XDestroyWindow(Tk_Display(foreignwinPtr->tkwin),
					foreignwinPtr->foreignWinId);
		    Tk_DeleteErrorHandler(handler);
		    XSync(Tk_Display(foreignwinPtr->tkwin), False);
		    hashEntryPtr = Tcl_FindHashEntry(&foreignWinTable,
					(char *) (foreignwinPtr->foreignWinId));
		    if (hashEntryPtr != NULL) {
			Tcl_DeleteHashEntry(hashEntryPtr);
		    }
		    foreignwinPtr->foreignWinId = None;
		}
		foreignwinPtr->tkwin = NULL;
		Tcl_DeleteCommandFromToken(foreignwinPtr->interp,
						foreignwinPtr->widgetCmd);
	    }
	    if (foreignwinPtr->updatePending) {
		Tcl_CancelIdleCall(ForeignWinDisplay,
					(ClientData) foreignwinPtr);
	    }
	    Tcl_EventuallyFree((ClientData) foreignwinPtr, ForeignWinDestroy);
	}
    } else if (eventPtr->type == ResizeRequest) {
	/*
	 * If we're supposed to follow the size of the foreign
	 * window, resize the Tk window accordingly.
	 */

#ifdef DEBUG
	fprintf(stderr, "Handling ResizeRequest event\n");
#endif
	if (eventPtr->xresizerequest.window == foreignwinPtr->foreignWinId
		&& foreignwinPtr->followSize) {
	    foreignwinPtr->reqWidth = eventPtr->xconfigurerequest.width;
	    foreignwinPtr->reqHeight = eventPtr->xconfigurerequest.height;
	    if (!foreignwinPtr->updatePending) {
		foreignwinPtr->updatePending = 1;
		Tcl_DoWhenIdle(ForeignWinDisplay, (ClientData) foreignwinPtr);
	    }
	}
    } else if (eventPtr->type == ReparentNotify) {
	/*
	 * Delete the Tk window if some other app takes the foreign
	 * window away. This is almost like getting a DestroyNotify
	 * event, except that we never delete the foreign window.
	 */

	if (eventPtr->xreparent.window == foreignwinPtr->foreignWinId
		&& eventPtr->xreparent.parent !=
				Tk_WindowId(foreignwinPtr->tkwin)) {
	    /* Foreign window was reparented; destroy the Tk window. */

	    hashEntryPtr = Tcl_FindHashEntry(&foreignWinTable,
				(char *) (foreignwinPtr->foreignWinId));
	    if (hashEntryPtr != NULL) {
		Tcl_DeleteHashEntry(hashEntryPtr);
	    }
	    foreignwinPtr->foreignWinId = None;
            Tcl_Preserve((ClientData) foreignwinPtr);
	    Tk_DestroyWindow(foreignwinPtr->tkwin);
	    foreignwinPtr->tkwin = NULL;
            Tcl_Release((ClientData) foreignwinPtr);
	}
    }
#ifdef DEBUG
else if (eventPtr->type == PropertyNotify) {
	/*
	 * If a PropertyNotify event occurred on the foreign
	 * window, synthesize one for the Tk window.
	 */

	XPropertyEvent *propertynotifyPtr = (XPropertyEvent *) eventPtr;
	fprintf(stderr,"Got PropertyNotify event for window 0x%x\n",
			propertynotifyPtr->window);
    }
#endif

#ifdef DEBUG
    fprintf(stderr, "Returning from ForeignWinEventProc\n");
#endif
}

/*
 *----------------------------------------------------------------------
 *
 * ForeignWinCmdDeletedProc --
 *
 *	This procedure is invoked when a widget command is deleted.  If
 *	the widget isn't already in the process of being destroyed,
 *	this command destroys it.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	The widget is destroyed.
 *
 *----------------------------------------------------------------------
 */

static void
ForeignWinCmdDeletedProc(clientData)
    ClientData clientData;	/* Pointer to widget record for widget. */
{
    ForeignWin *foreignwinPtr = (ForeignWin *) clientData;
    Tk_Window tkwin = foreignwinPtr->tkwin;

    /*
     * This procedure could be invoked either because the window was
     * destroyed and the command was then deleted (in which case tkwin
     * is NULL) or because the command was deleted, and then this procedure
     * destroys the widget.
     */

    if (tkwin != NULL) {
	foreignwinPtr->tkwin = NULL;
	Tk_DestroyWindow(tkwin);
    }
}

/*
 *--------------------------------------------------------------
 *
 * ForeignWinDisplay --
 *
 *	This procedure redraws the contents of a foreignwin window.
 *	It is invoked as a do-when-idle handler, so it only runs
 *	when there's nothing else for the application to do.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	The window gets resized, if necessary.
 *
 *--------------------------------------------------------------
 */

static void
ForeignWinDisplay(clientData)
    ClientData clientData;	/* Information about window. */
{
    ForeignWin *foreignwinPtr = (ForeignWin *) clientData;
    Tk_Window tkwin = foreignwinPtr->tkwin;
    XWindowAttributes winAtts;
    XWindowChanges changes;
    Status status;

#ifdef DEBUG
    fprintf(stderr, "In ForeignWinDisplay\n");
    fprintf(stderr, "	reqWidth = %1d, reqHeight = %1d\n",
		foreignwinPtr->reqWidth, foreignwinPtr->reqHeight);
#endif

    /* Don't do anything if we don't have to... */

    if (foreignwinPtr->reqWidth < 0 && foreignwinPtr->reqHeight < 0) {
	return;
    }

    Tcl_Preserve((ClientData) foreignwinPtr);

    /* Do we need to resize the Tk window? */

    if (Tk_Width(tkwin) != foreignwinPtr->reqWidth
		|| Tk_Height(tkwin) != foreignwinPtr->reqHeight) {
#ifdef DEBUG
	fprintf(stderr, "	Calling Tk_GeometryRequest\n");
#endif
	Tk_GeometryRequest(tkwin, foreignwinPtr->reqWidth,
					foreignwinPtr->reqHeight);

#if 1
	/* Give geometry managers a chance to do their thing. */

#ifdef DEBUG
	fprintf(stderr, "	Handling idle events\n");
#endif
	while (Tcl_DoOneEvent(TCL_IDLE_EVENTS)) {
	    /* Empty loop body. */
	}
#endif

    }

    /* Do we need to resize the foreign window? */

    status = XGetWindowAttributes(Tk_Display(tkwin),
				foreignwinPtr->foreignWinId, &winAtts);
    if (winAtts.width != foreignwinPtr->reqWidth
		|| winAtts.height != foreignwinPtr->reqHeight) {
#ifdef DEBUG
	fprintf(stderr, "	Calling XConfigureWindow\n");
#endif
	changes.x = 0;
	changes.y = 0;
	changes.width = foreignwinPtr->reqWidth;
	changes.height = foreignwinPtr->reqHeight;
	XConfigureWindow(Tk_Display(foreignwinPtr->tkwin),
			foreignwinPtr->foreignWinId,
			CWX|CWY|CWWidth|CWHeight, &changes);
    }


    /* Make sure we don't redo things... */

    foreignwinPtr->updatePending = 0;
    foreignwinPtr->reqWidth = -1;
    foreignwinPtr->reqHeight = -1;

    Tcl_Release((ClientData) foreignwinPtr);

#ifdef DEBUG
    fprintf(stderr, "Done with ForeignWinDisplay\n");
#endif
}

/*
 *----------------------------------------------------------------------
 *
 * ForeignWinDestroy --
 *
 *	This procedure is invoked by Tcl_EventuallyFree or Tcl_Release
 *	to clean up the internal structure of a foreignwin at a safe time
 *	(when no-one is using it anymore).
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Everything associated with the foreignwin is freed up.
 *
 *----------------------------------------------------------------------
 */

static void
ForeignWinDestroy(memPtr)
    char *memPtr;		/* Info about foreignwin widget. */
{
    ForeignWin *foreignwinPtr = (ForeignWin *) memPtr;
    Tcl_HashEntry *hashEntryPtr;
    Tk_ErrorHandler handler;
    int gotXError;

    /* Destroy the foreign window if it hasn't been destroyed already. */

    if (foreignwinPtr->foreignWinId != None) {
	hashEntryPtr = Tcl_FindHashEntry(&foreignWinTable,
				(char *) (foreignwinPtr->foreignWinId));
	if (hashEntryPtr != NULL) {
	    if (foreignwinPtr->tkwin != NULL) {
		gotXError = 0;
		handler = Tk_CreateErrorHandler(Tk_Display(foreignwinPtr->tkwin),
			-1, -1, -1,
			Tkwm_XErrorHandler, (ClientData) (&gotXError));
		XDestroyWindow(Tk_Display(foreignwinPtr->tkwin),
					foreignwinPtr->foreignWinId);
		Tk_DeleteErrorHandler(handler);
		XSync(Tk_Display(foreignwinPtr->tkwin), False);
	    }
	    Tcl_DeleteHashEntry(hashEntryPtr);
	}
	foreignwinPtr->foreignWinId = None;
    }

    Tk_DeleteOptionTable(foreignwinPtr->optionTable);
    ckfree((char *) foreignwinPtr);
}

int
Foreignwin_Init(interp)
    Tcl_Interp *interp;
{
    if(Tcl_PkgRequire(interp, "Tk", "8.4", 0) == NULL) {
        return TCL_ERROR;
    }

    Tcl_CreateObjCommand(interp, "foreignwin", Tk_ForeignWinCmd, Tk_MainWindow(interp), NULL);
    Tcl_PkgProvide(interp, "Foreignwin", VERSION);

    return TCL_OK;
}

